@@ -2,14 +2,17 @@ import time
 import inspect
 import asyncio
 import datetime
+import logging
 from typing import Union, Callable, Awaitable, Tuple, Optional, ClassVar
 
 from serial_json import DataClass, field, field_property, MISSING, \
     Weekdays, weekdays_property, weekdays_attr_property, \
     datetime_property, time_property, timedelta_attr_property, seconds_property, make_datetime
 
+from .utils import call, call_async
 
-__all__ = ['Schedule']
+
+__all__ = ['Schedule', 'RepeatSchedule']
 
 
 class Schedule(DataClass):
@@ -77,30 +80,33 @@ class Schedule(DataClass):
         last_run (DateTime/str)[None]: Date and time to make the next_run from.
         next_run (DateTime/str)[None]: Manually set the next run time.
     """
-    days: int = 0
-    hours: int = 0
-    minutes: int = 0
-    milliseconds: int = 0
-    microseconds: int = 0
-    seconds: Union[int, float] = seconds_property('seconds')
-    weeks: int = 0
-    interval: datetime.timedelta = timedelta_attr_property(required=False)
+    days: int = field(0, skip_repr=0, skip_dict=0)
+    hours: int = field(0, skip_repr=0, skip_dict=0)
+    minutes: int = field(0, skip_repr=0, skip_dict=0)
+    milliseconds: int = field(0, skip_repr=0, skip_dict=0)
+    microseconds: int = field(0, skip_repr=0, skip_dict=0)
+    seconds: Union[int, float] = seconds_property('seconds', skip_repr=0, skip_dict=0)
+    weeks: int = field(0, skip_repr=0, skip_dict=0)
+    interval: datetime.timedelta = timedelta_attr_property(required=False, repr=False, dict=False)
 
-    weekdays: Weekdays = weekdays_property('weekdays', default_factory=Weekdays)
-    sunday: bool = weekdays_attr_property('weekdays', 'sunday')
-    monday: bool = weekdays_attr_property('weekdays', 'monday')
-    tuesday: bool = weekdays_attr_property('weekdays', 'tuesday')
-    wednesday: bool = weekdays_attr_property('weekdays', 'wednesday')
-    thursday: bool = weekdays_attr_property('weekdays', 'thursday')
-    friday: bool = weekdays_attr_property('weekdays', 'friday')
-    saturday: bool = weekdays_attr_property('weekdays', 'saturday')
+    weekdays: Weekdays = weekdays_property('weekdays', default_factory=Weekdays,
+                                           skip_repr=Weekdays(), skip_dict=Weekdays())
+    sunday: bool = weekdays_attr_property('weekdays', 'sunday', repr=False, dict=False)
+    monday: bool = weekdays_attr_property('weekdays', 'monday', repr=False, dict=False)
+    tuesday: bool = weekdays_attr_property('weekdays', 'tuesday', repr=False, dict=False)
+    wednesday: bool = weekdays_attr_property('weekdays', 'wednesday', repr=False, dict=False)
+    thursday: bool = weekdays_attr_property('weekdays', 'thursday', repr=False, dict=False)
+    friday: bool = weekdays_attr_property('weekdays', 'friday', repr=False, dict=False)
+    saturday: bool = weekdays_attr_property('weekdays', 'saturday', repr=False, dict=False)
 
     repeat: bool = False
-    at: Optional[datetime.time] = time_property('at', allow_none=True, required=False)
-    start_on: Optional[datetime.datetime] = datetime_property('start_on', default_factory=datetime.datetime.now)
-    end_on: Optional[datetime.datetime] = datetime_property('end_on', allow_none=True, required=False)
-    last_run: Optional[datetime.datetime] = datetime_property('last_run', allow_none=True, required=False)
-    _next_run: Union[datetime.datetime, None] = field(default=None, repr=False)
+    at: datetime.time = time_property('at', allow_none=True, required=False, skip_repr=None, skip_dict=None)
+    start_on: datetime.datetime = datetime_property('start_on', default_factory=datetime.datetime.now, repr=False)
+    end_on: datetime.datetime = datetime_property('end_on', allow_none=True, required=False, skip_repr=None, skip_dict=None)
+    last_run: datetime.datetime = datetime_property('last_run', allow_none=True, required=False, repr=False, skip_dict=None)
+    _next_run: Union[datetime.datetime, None] = field(default=None, repr=False, skip_dict=None)
+
+    logger: logging.Logger = field(default=logging.getLogger('asyncio'), repr=False, dict=False, hash=False, compare=False)
 
     @field_property(default=None)
     def next_run(self) -> Union[datetime.datetime, None]:
@@ -180,17 +186,26 @@ class Schedule(DataClass):
         self.wait()
         self.reschedule()
 
-        if callable(callback):
-            return callback(*args, **kwargs)
+        task = asyncio.Task.current_task().get_name()
+        self.logger.info(f'Running Task "{task}" with {self}')
+
+        try:
+            return call(callback, *args, **kwargs)
+        except Exception as err:
+            self.logger.critical(f'Error in Task "{task}": {err}')
 
     async def call_async(self, callback: Callable[..., Awaitable[None]] = None, *args, **kwargs) -> object:
         """Run the set callback and setup repeat if set."""
         await self.wait_async()
         await self.reschedule_async()
-        if inspect.iscoroutinefunction(callback):
-            return await callback(*args, **kwargs)
-        elif callable(callback):
-            return callback(*args, **kwargs)
+
+        task = asyncio.Task.current_task().get_name()
+        self.logger.info(f'Running Task "{task}" with {self}')
+
+        try:
+            return await call_async(callback, *args, **kwargs)
+        except Exception as err:
+            self.logger.critical(f'Error in Task "{task}": {err}')
 
     def run(self, callback: Callable = None, *args, **kwargs) -> 'Schedule':
         """Loop until and call this function until the schedule ends."""
@@ -204,10 +219,6 @@ class Schedule(DataClass):
             await self.call_async(callback, *args, **kwargs)
         return self
 
-    WEEKDAY_ATTRS = tuple(Weekdays.DAYS)
-    INTERVAL_ATTRS = ('weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds')
-    STRING_NAME_ATTRS = INTERVAL_ATTRS + ('repeat', 'at')
-
     def allowed_weekdays(self) -> Tuple[str]:
         """Return if all of the days of the week are None."""
         return tuple(self.weekdays)
@@ -253,14 +264,6 @@ class Schedule(DataClass):
 
         return dt
 
-    def __str__(self) -> str:
-        attrs = tuple('{}={}'.format(attr, getattr(self, attr))
-                      for attr in self.STRING_NAME_ATTRS if getattr(self, attr))
 
-        allowed_weekdays = self.allowed_weekdays()
-        if allowed_weekdays != self.WEEKDAY_ATTRS:
-            attrs = attrs + ('on=[' + ', '.join(allowed_weekdays) + ']',)
-        return '{cls}({attrs})'.format(cls=self.__class__.__name__, attrs=', '.join(attrs))
-
-    def __repr__(self) -> str:
-        return '<{str} at {id}>'.format(str=self.__str__(), id=id(self))
+class RepeatSchedule(Schedule):
+    repeat: bool = True
